/*!***************************************************************************
*! FILE NAME  : audio_thr.c
*! DESCRIPTION: It is implementation of thread, which captured data from audio
*! device and dump it to audio track Matroska container.
*! Timestamp data synced with video frames
*! Copyright (C) 2010 Elphel, Inc.
*! -----------------------------------------------------------------------------**
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!
*!  $Log$
*/
#define _GNU_SOURCE
#define __USE_POSIX199506 
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>
#include <math.h>

#include <malloc.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <time.h>
#include <alsa/asoundlib.h>
#include <asm/byteorder.h>
#include <asm/elphel/c313a.h>

#include "defines.h"

//Timestamp GetRelativeTime(void);

static snd_pcm_sframes_t (*readi_func)(snd_pcm_t *handle, void *buffer, snd_pcm_uframes_t size);
//static snd_pcm_sframes_t (*readn_func)(snd_pcm_t *handle, void **bufs, snd_pcm_uframes_t size);

snd_pcm_hw_params_t 	*hw_params;
snd_pcm_sw_params_t 	*swparams;
static unsigned 	buffer_time = 0;
static snd_pcm_uframes_t chunk_size = 0;   // 48000* 20000(20ms)/1000000 = 960 frames
static size_t 		bits_per_sample, 
			bits_per_frame;
static size_t 		chunk_bytes;
uint64_t   		atime_stamp =0;
unsigned long		rtc[2];



// Functions for dumping audio frames to various formats container

int a_movwrite(void)
{

	//quicktime_insert_keyframe(state->qtfile, state->frameno, 0);
	//quicktime_write_audio(state->qtfile, state->audiobuf, chunk_size, 0);

return 0;	
}


/** Audio thread for capture audio from sound card
 */
void * thread_a(void *arg)
{
	int32_t		r;
	int		err;

        pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	usleep(20000);	

	while(!(state->done)) {
		DBG(3,"Waiting to start audio capture ...\n");
		pthread_mutex_lock (&thread_lock);
		pthread_mutex_unlock (&thread_lock);
		if(state->done)
			break;
		DBG(3,"starting audio capture ...\n");
		if(state->audio_enabled) {
			if ((err = snd_pcm_prepare (state->capture_handle)) < 0) {
				DBG(0, "cannot prepare audio interface for use (%s)\n", snd_strerror (err));
				state->can_capture=0;
			} else 	{
				// audio card started - get diff time between this trhead and video thread and
				// correct timestamp for audio stream
				state->can_capture=1;
				get_fpga_time(rtc);
				// correct timestamp(used in Matroska) for syncing with video,
				// video and audio thread can't execute in same time - differences between starting
				// used for correction
				state->atime_start=(uint64_t)rtc[1]+(uint64_t)rtc[0]*1000000l;
			        DBG(3, "rtc[1]=%ld rtc[0]=%ld %lld\n",rtc[1],rtc[0],state->atime_start);	
				//atime_stamp=state->atime_start -
				//	(state->frame_params.timestamp_usec+1000000*state->frame_params.timestamp_sec);
			}
			DBG(3,"started audio capture\n");

			while(state->can_capture) {
				if(!state->synced) {
					usleep(10);
					continue;
				}
				DBG(4,"reading captured frames...\n");
				r = readi_func(state->capture_handle, state->audiobuf, chunk_size);
				if (r == -EAGAIN) {
					usleep(10);
					continue;
				}
				if(r<0) {
					DBG(3,"status frames read r=%d chunk_bytes=%d ",r, chunk_bytes );
					err = snd_pcm_state(state->capture_handle);
					DBG(0,"audio error: %s\n", snd_strerror(err));
				}
				if(r>0) {
					DBG(3,"Writing audio data\n");
					// sync writing to MKV file with video thread	
					pthread_mutex_lock (&mkv_lock);
					a_write_func();	
					pthread_mutex_unlock (&mkv_lock);
				}
				if(state->a_stopping)
					break;
				continue;
			}
			if(state->can_capture) {
				pthread_mutex_lock (&stopping_lock); 
				DBG(3,"Stopping audio card ...\n");	
				if((err=snd_pcm_drop(state->capture_handle)) <0) {
					DBG(3, "cannot stop audio interface. Err: (%s)\n", snd_strerror (err));
				}
				state->a_stopping=2;
				pthread_cond_signal(&cond);
				DBG(3,"Stopped audio...\n");	
				pthread_mutex_unlock (&stopping_lock);
			}
		} else {
			while(1) {
				usleep(COMMAND_LOOP_DELAY);
				if(state->a_stopping)
					break;
			}
			pthread_mutex_lock (&stopping_lock); 
			DBG(3,"Stopping audio worker...\n");	
			state->a_stopping=2;
			pthread_cond_signal(&cond);
			pthread_mutex_unlock (&stopping_lock);
		}

		usleep(20);
	}

DBG(2,"End of audio thread.\n");
return NULL;
}


void set_audio_pars(void)
{
	state->format = DEFAULT_AUDIO_FORMAT; //SND_PCM_FORMAT_S16_LE
	cammkv_set_audio_volume(AUDIO_DEF_VOLUME);
	cammkv_set_audio_format("48000/2");
	strcpy(state->pcm_name,"plug:hw:0");
	state->period_time = 20000;    // period for capture frames 20 ms 
	state->buffer_time = 5000000;  // audio buffer time 5 s
}

int init_audio(void)
{
	snd_pcm_uframes_t buffer_size;
	int err;
	size_t n;
	unsigned int rate;
	//snd_pcm_uframes_t start_threshold, stop_threshold;

	state->synced=0;	
	if(!state->audio_enabled)
		return 0;
	set_audio_pars();
	DBG(3,"Starting init audio subsystem\n");
	if (state->mmap_flag) {
		readi_func = snd_pcm_mmap_readi;
		//readn_func = snd_pcm_mmap_readn;
	} else {
		readi_func = snd_pcm_readi;
		//readn_func = snd_pcm_readn;
	}
	//if ((err = snd_pcm_open (&state->capture_handle, "plug:hw:0", SND_PCM_STREAM_CAPTURE, 0)) < 0) {
	if ((err = snd_pcm_open (&state->capture_handle, "plug:hw:0", SND_PCM_STREAM_CAPTURE, SND_PCM_NONBLOCK)) < 0) {
		DBG(0, "cannot open audio device %s (%s)\n", state->pcm_name, 	snd_strerror (err));
		return -1;
	}
	
	snd_pcm_hw_params_alloca(&hw_params);
	snd_pcm_sw_params_alloca(&swparams);
	if((err = snd_pcm_hw_params_any(state->capture_handle, hw_params)) <0) {
		DBG(0,"Broken configuration for this PCM: no configurations available");
		return -1;
	}
	if (state->mmap_flag) {
		snd_pcm_access_mask_t *mask = alloca(snd_pcm_access_mask_sizeof());
		snd_pcm_access_mask_none(mask);
		snd_pcm_access_mask_set(mask, SND_PCM_ACCESS_MMAP_INTERLEAVED);
		err = snd_pcm_hw_params_set_access_mask(state->capture_handle, hw_params, mask);
	} else 
		err = snd_pcm_hw_params_set_access(state->capture_handle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED);
	if (err < 0) {
		DBG(0,"Access type to sound card not available");
		return -1;
	}

	if((err = snd_pcm_hw_params_set_format(state->capture_handle, hw_params, state->format))<0) {
		DBG(0,"Sample format non available");
		return -1;
	}
	rate = state->rate;
	if((err = snd_pcm_hw_params_set_rate_near(state->capture_handle, hw_params, &state->rate, 0))<0) {
		DBG(0,"Can't set sample rate [%d]\n",state->rate);
		return -1;
	}
	DBG(3,"Audio rate are %d\n",state->rate);
	if ((float)rate * 1.05 < (float)state->rate || (float)rate * 0.95 > (float)state->rate) {
			char plugex[64];
			const char *pcmname = snd_pcm_name(state->capture_handle);
			DBG(1,"Warning: rate is not accurate (requested = %iHz, got = %iHz)\n", rate, state->rate);
			if (! pcmname || strchr(snd_pcm_name(state->capture_handle), ':'))
				*plugex = 0;
			else
				snprintf(plugex, sizeof(plugex), "(-Dplug:%s)", snd_pcm_name(state->capture_handle));
			DBG(0,"please, try the plug plugin %s\n",plugex);
	}

	rate = state->rate;
	if((err = snd_pcm_hw_params_set_channels(state->capture_handle, hw_params, state->channels))<0) {
		DBG(0,"Channels count non available");
		return -1;
	}
	if((err = snd_pcm_hw_params_get_buffer_time_max(hw_params,  &buffer_time, 0))<0) {
		DBG(0,"Can't get buffer time max.\n");
		return -1;
	}
	DBG(3,"Max buffer time=%d\n",buffer_time);
	if (buffer_time > state->buffer_time)
		buffer_time = state->buffer_time;
	buffer_time = 5000000;
	
	//state->period_time = buffer_time / 4;
	if((err = snd_pcm_hw_params_set_period_time_near(state->capture_handle, hw_params, &state->period_time, 0))<0) {
		DBG(0,"Can't set period time\n");
		return -1;
	}
	DBG(3,"period time=%d\n",state->period_time);
	if((err = snd_pcm_hw_params_set_buffer_time_near(state->capture_handle, hw_params, &buffer_time, 0))<0) {
		DBG(0,"Can't set buffer time\n");
		return -1;
	}
	DBG(3,"buffer time near =%d\n",buffer_time);
	
	err = snd_pcm_hw_params(state->capture_handle, hw_params);
	if (err < 0) {
		DBG(0,"Unable to install HW params: %s\n",snd_strerror (err));
		return -1;
	}

	snd_pcm_hw_params_get_period_size(hw_params, &chunk_size, 0);
	snd_pcm_hw_params_get_buffer_size(hw_params, &buffer_size);
	if (chunk_size == buffer_size) {
		DBG(0,"Can't use period equal to buffer size (%lu == %lu)", chunk_size, buffer_size);
		return -1;
	}
	DBG(3,"chunk size(in frames) =%d\n",(int)chunk_size);
	snd_pcm_sw_params_current(state->capture_handle, swparams);
	if(state->mmap_flag)
		n = 4; 
	else    
		n = chunk_size;
	err = snd_pcm_sw_params_set_avail_min(state->capture_handle, swparams, n);
	if(err<0) {
		DBG(0,"Can't set sw avail min.\n");
		return -1;
	}
	/* round up to closest transfer boundary */
/*	n = buffer_size;
	start_threshold = n;
	if((err = snd_pcm_sw_params_set_start_threshold(state->capture_handle, swparams, start_threshold)) <0) {
		DBG(0,"Can't set sw start threshold.\n");
		return -1;
	}
	stop_threshold = buffer_size + (double) rate * stop_delay / 1000000;
	if((err = snd_pcm_sw_params_set_stop_threshold(state->capture_handle, swparams, stop_threshold))<0) {
		DBG(0,"Can't set sw stop threshold.\n");
		return -1;
	}

	if (snd_pcm_sw_params(state->capture_handle, swparams) < 0) {
		DBG(0,"unable to install sw params");
		return -1;
	}
*/
	bits_per_sample = snd_pcm_format_physical_width(state->format);
	state->bitDepth = bits_per_sample;
	DBG(3,"bits per sample=%d\n",bits_per_sample);
	bits_per_frame = bits_per_sample * state->channels;
	chunk_bytes = chunk_size * bits_per_frame / 8;
	DBG(3,"Audio chunk size(bytes) %d \n",chunk_bytes);
	state->audiobuf = (uint8_t *)malloc(chunk_bytes);
	state->chunk_bytes = chunk_bytes;
	if (state->audiobuf == NULL) {
		DBG(0,"not enough memory for audio buffer");
		return -1;
	}
	DBG(3,"Audio subsytem inited, ");
	//state->can_capture=1; // set to 1 if sound card opened success in main audio thread
	atime_stamp=0;
	DBG(3,"audio can capture!\n");
return 0;
}

int deinit_audio(void)
{
	state->can_capture = 0;
	if(state->audio_enabled) {
	DBG(3,"Close audio capture handler\n");
		snd_pcm_close (state->capture_handle);
	}
	if(state->audiobuf != NULL) {
		DBG(3,"Freeing audio buffer\n");
		free(state->audiobuf);
		state->audiobuf = NULL;
	}

return (0);
}

int stopping_audio(void)
{
	state->a_stopping=1;
	pthread_mutex_lock (&stopping_lock);
	while(state->a_stopping==1)
		pthread_cond_wait(&cond,&stopping_lock);
	state->a_stopping=0;
	pthread_mutex_unlock (&stopping_lock);

return 0;
}
