/*!***************************************************************************
*! FILE NAME  : cammkv_utl.c
*! DESCRIPTION: Support functions for cammkv utility.
*! Copyright (C) 2010 Elphel, Inc.
*! -----------------------------------------------------------------------------**
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!
*!  $Log$
*/
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <inttypes.h>
#include <pthread.h>
#include <string.h>
#include <sys/time.h>

#include "config.h"
#include "defines.h"

typedef struct mk_Writer_s {
        FILE *fp;
        uint64_t f_pos;
        uint64_t f_eof;

        int64_t duration_ptr;
        int64_t seekhead_ptr;
        int64_t segment_ptr;
        int64_t segmentuid_ptr;
} _mk_Writer_s_t;

extern uint64_t atime_stamp;

static int GPWrite(unsigned long *data, int length) 
{
	return write(state->fd_fparmsall, (void *)data, length);
}

/**
 * @brief Read either G_* parameter (these are 'regular' values defined by number) or P_* parameter
 *         (it can be read for up to 6 frames ahead, but current interface only allows to read last/current value)
 * @param GPNumber parameter number (as defined in c313a.h), G_* parameters have numbers above FRAMEPAR_GLOBALS, P_* - below)
 * @return parameter value
 */
unsigned long getGPValue(unsigned long GPNumber) {
	return (GPNumber>=FRAMEPAR_GLOBALS)?
		GLOBALPARS(GPNumber):
		framePars[GLOBALPARS(G_THIS_FRAME) & PARS_FRAMES_MASK].pars[GPNumber];
}

/**
 * @brief Set value of the specified global (G_*) parameter
 * @param GNumber - parameter number (as defined in c313a.h)
 * @param value  - value to set
 */
void setGValue(unsigned long  GNumber,   unsigned long value) {
    GLOBALPARS(GNumber)=value;
}


void get_fpga_time(unsigned long *rtc) 
{
	unsigned long write_data[6];

	write_data[0] = FRAMEPARS_GETFPGATIME;
	write_data[1] = 0;
	GPWrite(write_data, sizeof(unsigned long) * 2);
	rtc[0] = getGPValue(G_SECONDS);
	rtc[1] = getGPValue(G_MICROSECONDS);
}



/////// MOV functions open/write/close for container ////////////////////////////////////////
#if 0
int dvr_movopen(void)
{
	DBG(3,"Trying open MOV file\n");
	sprintf(state->path,"%s%010ld_%06ld.mov",state->path_prefix,
			state->this_frame_params.timestamp_sec,state->this_frame_params.timestamp_usec);
	DBG(3,"MOV: path= %s audio_enabled=%d\n", state->path,state->audio_enabled);

        state->qtfile = quicktime_open(state->path, 0, 1);
	if( !(state->qtfile) )
	{
		DBG(0,"Could not create output MOV file: \n" );
		return -1;
	}
	DBG(3,"QuickTime output file opened, trying add audio/video track %d %d \n",state->width, state->height);
	// First - add audio track!!!! if audio enabled
	DBG(3,"Set MOV audio parameters, if enabled(%d)\n",state->audio_enabled);
	if(state->audio_enabled) {
		DBG(3,"Set MOV audio parameters\n");
		quicktime_set_audio(state->qtfile, state->channels, state->rate, state->bitDepth, "lpcm");  // PCM audio
	}
	/* Now add Video track */
	state->fps=25.1; // set default FPS, value after dot must not zero - best for VFR, 
			 // FPS recalculated for each frame - need for syncing A/V
	quicktime_set_video(state->qtfile, 1, state->width, state->height, state->fps, "jpeg");

return 0;
}

int dvr_movclose(void)
{
	DBG(3,"Closing MOV file\n");
	quicktime_close(state->qtfile);
        state->qtfile=NULL;	
return 0;
}

#endif
//////////////////////////////////////////////////////////////////////////////////////////////////


int cammkv_free(void)
{

return 0;
}


void  cammkv_status(char * fn, int xml) 
{
	long 	_len=0;
	int 	_dur,_udur;
	//TODO:make it XML file
	FILE	*f;
	char 	*_state, *_output_format, *_using_exif, 
		*_using_global_pointer, *_compressor_state;
	int 	_b_free, _b_used; // , save_p;
	int 	_frames_remain=0;
	int 	_sec_remain=0;
	int 	_frames_skip=0;
	int 	_sec_skip=0;
	char 	*_kml_enable, *_kml_used, *_kml_height_mode;
	char 	*_audio_enable;

	_kml_enable=      	state->kml_enable?"yes":"no";
	_kml_used=        	state->kml_used?"yes":"no";
	_kml_height_mode= 	state->kml_height_mode?"GPS altitude":"map ground level";  //! 1 - actual, 0 - ground
	_audio_enable=  	state->audio_enabled?"yes":"no";
	//_allow_sync=      	state->set_allow_sync?"yes":"no";


	_b_free = getGPValue(G_FREECIRCBUF);
	_b_used = getGPValue(G_CIRCBUFSIZE)-getGPValue(G_FREECIRCBUF);

	if (!fn) f=stdout;
	else if (strcmp(fn, "stdout")==0) f=stdout;
	else if (strcmp(fn, "stderr")==0) f=stderr;
	else {
		if (!((f=fopen (fn,"w")))) {
			DBG(0,"Error opening %s\n", fn);
			return;
		}
	}
	if(state->mkv_file) { 
		_len=((_mk_Writer_s_t*)state->mkv_file)->f_pos;
		//pthread_mutex_lock (&stopping_lock); 
		//_len=ftell(((_mk_Writer_s_t*)state->mkv_file)->fp);
		//DBG(3,"------->f_pos=%d\n",((_mk_Writer_s_t*)state->mkv_file)->f_pos); 
		//pthread_mutex_unlock (&stopping_lock); 
	}
	_dur= state->this_frame_params.timestamp_sec-state->frame_params.timestamp_sec;
	_udur=state->this_frame_params.timestamp_usec-state->frame_params.timestamp_usec;
	if (_udur<0) {
		_dur-=1;
		_udur+=1000000;
	} else if (_udur>=1000000) {
		_dur+=1;
		_udur-=1000000;
	}
	_state= state->running? "running":(state->starting?"starting":"stopped");
	_output_format = state->video_format?((state->format==CAMMKV_FORMAT_OGM)?"ogm":
			((state->video_format==CAMMKV_FORMAT_JPEG)?"jpeg":
			((state->video_format==CAMMKV_FORMAT_MOV)?"mov":
			((state->video_format==CAMMKV_FORMAT_MKV)?"mkv":
			 "other")))):"none";
	_using_exif = state->exif?"yes":"no";
	_using_global_pointer = state->save_gp?"yes":"no";
	_compressor_state = (getGPValue(P_COMPRESSOR_RUN)==2)?"running":"stoppped";
	if ( state->frames_skip >0 ) {
		_frames_remain = state->frames_skip_left;
		_frames_skip   = state->frames_skip;
	} else if ( state->frames_skip <0 ) {
		_sec_remain= (state->frames_skip_left - state->this_frame_params.timestamp_sec);
		_sec_skip=-(state->frames_skip);
	}


	if (xml) {
		fprintf (f,"<?xml version=\"1.0\"?>\n" \
				"<cammkv_state>\n" \
				"  <state>\"%s\"</state>\n" \
				"  <compressor_state>\"%s\"</compressor_state>\n" \
				"  <file_name>\"%s\"</file_name>\n" \
				"  <frame_number>%d</frame_number>\n" \
				"  <file_duration>%d.%06d</file_duration>\n" \
				"  <file_length>%ld</file_length>\n" \
				"  <frame_period>%d</frame_period>\n" \
				"  <frames_skip>%d</frames_skip>\n" \
				"  <seconds_skip>%d</seconds_skip>\n" \
				"  <frames_skip_left>%d</frames_skip_left>\n" \
				"  <seconds_skip_left>%d</seconds_skip_left>\n" \
				"  <frame_width>%d</frame_width>\n" \
				"  <frame_height>%d</frame_height>\n" \
				"  <format>\"%s\"</format>\n" \
				"  <exif>\"%s\"</exif>\n" \
				"  <prefix>\"%s\"</prefix>\n" \
				"  <max_duration>%d</max_duration>\n" \
				"  <max_length>%d</max_length>\n" \
				"  <max_frames>%d</max_frames>\n" \
				"  <timescale>%f</timescale>\n" \
				"  <buffer_overruns>%d</buffer_overruns>\n" \
				"  <buffer_minimal>%d</buffer_minimal>\n" \
				"  <buffer_free>%d</buffer_free>\n" \
				"  <buffer_used>%d</buffer_used>\n" \
				"  <circbuf_rp>%d</circbuf_rp>\n" \
				"  <debug_output>\"%s\"</debug_output>\n" \
				"  <debug_level>%d</debug_level>\n" \
				"  <use_global_rp>\"%s\"</use_global_rp>\n" \
				"  <kml_enable>\"%s\"</kml_enable>\n" \
				"  <kml_used>\"%s\"</kml_used>\n" \
				"  <kml_path>\"%s\"</kml_path>\n" \
				"  <kml_horHalfFov>\"%f\"</kml_horHalfFov>\n" \
				"  <kml_vertHalfFov>\"%f\"</kml_vertHalfFov>\n" \
				"  <kml_near>\"%f\"</kml_near>\n" \
				"  <kml_height_mode>\"%s\"</kml_height_mode>\n" \
				"  <kml_height>\"%f\"</kml_height>\n" \
				"  <kml_period>%d</kml_period>\n" \
				"  <kml_last_ts>%d.%06d</kml_last_ts>\n" \
				"  <audio_enable>\"%s\"</audio_enable>\n" \
				"  <audio_channels>\"%d\"</audio_channels>\n" \
				"  <audio_rate>\"%d\"</audio_rate>\n" \
				"  <audio_volume>\"%d\"</audio_volume>\n" \
				"</cammkv_state>\n",
				_state,_compressor_state,state->path,state->frameno,_dur,_udur,_len,state->frame_period, \
				_frames_skip,_sec_skip,_frames_remain, _sec_remain, \
				state->width,state->height,_output_format,_using_exif, \
				state->path_prefix, state->segment_duration, state->segment_length, state->max_frames, state->timescale, \
				state->buf_overruns, state->buf_min, _b_free, _b_used, state->cirbuf_rp, \
				state->debug_name, debug_level, _using_global_pointer, \
				_kml_enable,_kml_used,state->kml_path,state->kml_horHalfFov,state->kml_vertHalfFov,state->kml_near,\
				_kml_height_mode,state->kml_height,state->kml_period,state->kml_last_ts,state->kml_last_uts, \
				_audio_enable,state->channels,state->rate,state->audio_volume);
	} else {
		fprintf (f,"state              %s\n",        _state);
		fprintf (f,"compressor state   %s\n",        _compressor_state);
		fprintf (f,"file               %s\n",        state->path);
		fprintf (f,"frame              %d\n",        state->frameno);
		fprintf (f,"file duration      %d.%06d sec\n",_dur,_udur);
		fprintf (f,"file length        %ld B\n",      _len);
		fprintf (f,"frame period       %d (0x%x)\n", state->frame_period,state->frame_period);
		if ( _frames_skip >0 ) fprintf (f,"frames to skip   %d (left %d)\n",_frames_skip, _frames_remain);
		if ( _sec_skip    <0 ) fprintf (f,"timelapse period  %d sec (remaining %d sec)\n", _sec_skip, _sec_remain);
		fprintf (f,"width              %d (0x%x)\n", state->width,state->width);
		fprintf (f,"height             %d (0x%x)\n", state->height,state->height);
		fprintf (f,"\n");
		fprintf (f,"output format      %s\n",        _output_format);
		fprintf (f,"using exif         %s\n",        _using_exif);
		fprintf (f,"path prefix:       %s\n",        state->path_prefix);
		fprintf (f,"max file duration: %d sec\n",    state->segment_duration);
		fprintf (f,"max file length:   %d B\n",      state->segment_length);
		fprintf (f,"max frames         %d\n",        state->max_frames);
		fprintf (f,"timescale          %f\n",        state->timescale);
		fprintf (f,"\n");
		fprintf (f,"buffer overruns    %d\n",        state->buf_overruns);
		fprintf (f,"buffer minimal     %d\n",        state->buf_min);
		fprintf (f,"buffer free        %d\n",        _b_free);
		fprintf (f,"buffer used        %d\n",        _b_used);
		fprintf (f,"circbuf_rp         %d (0x%x)\n", state->cirbuf_rp,state->cirbuf_rp);
		fprintf (f,"\n");
		fprintf (f,"debug output to    %s\n",        state->debug_name);
		fprintf (f,"debug level        %d\n",        debug_level);
		fprintf (f,"use global pointer %s\n",        _using_global_pointer);
		fprintf (f,"\n\n");
		fprintf (f,"kml_enable         %s\n",        _kml_enable);
		fprintf (f,"kml_used           %s\n",        _kml_used);
		fprintf (f,"kml_path           %s\n",        state->kml_path);
		fprintf (f,"kml_horHalfFov     %f degrees\n",state->kml_horHalfFov);
		fprintf (f,"kml_vertHalfFov    %f degrees\n",state->kml_vertHalfFov);
		fprintf (f,"kml_near           %f m\n",      state->kml_near);
		fprintf (f,"kml height mode    %s\n",        _kml_height_mode);
		fprintf (f,"kml_height (extra) %f m\n",      state->kml_height);
		fprintf (f,"kml_period         %d\n",        state->kml_period);
		fprintf (f,"kml_last_ts        %d.%06d\n",   state->kml_last_ts,state->kml_last_uts);
		fprintf (f,"\n\n");
		fprintf (f,"audio_enable       %s\n",        _audio_enable);
		fprintf (f,"audio_channels     %d\n",        state->channels);
		fprintf (f,"audio_rate         %d\n",        state->rate);
		fprintf (f,"\n\n");

	}
	if ((f!=stdout) && (f!=stderr)) fclose (f);
	if (state->buf_overruns>=0) state->buf_overruns=0; //! resets overruns after reading status , so "overruns" means since last reading status
	state->buf_min=_b_free;
}

